# -*- coding: future_fstrings -*-
import subprocess
import os
import sys
from io import StringIO
from stimela import utils
import json
import stimela
import time
import datetime
import tempfile


class DockerError(Exception):
    pass


def build(image, build_path, tag=None, build_args=None, fromline=None, args=[]):
    """ build a podman image"""

    if tag:
        image = ":".join([image, tag])

    bdir = tempfile.mkdtemp()
    os.system('cp -r {0:s}/* {1:s}'.format(build_path, bdir))
    if build_args:
        stdw = tempfile.NamedTemporaryFile(dir=bdir, mode='w')
        with open("{}/Dockerfile".format(bdir)) as std:
            dfile = std.readlines()

        for line in dfile:
            if fromline and line.lower().startswith('from'):
                stdw.write('FROM {:s}\n'.format(fromline))
            elif line.lower().startswith("cmd"):
                for arg in build_args:
                    stdw.write(arg+"\n")
                stdw.write(line)
            else:
                stdw.write(line)
        stdw.flush()
        utils.xrun("podman build", args+["--force-rm", "--no-cache", "-f", stdw.name,
                                         "-t", image,
                                         bdir])

        stdw.close()
    else:
        utils.xrun("podman build", args+["--force-rm", "--no-cache", "-t", image,
                                         bdir])

    os.system('rm -rf {:s}'.format(bdir))


def pull(image, tag=None, force=False):
    """ pull a podman image """
    if tag:
        image = ":".join([image, tag])

    utils.xrun("podman", ["pull", "docker.io/"+image])


def seconds_hms(seconds):
    return str(datetime.timedelta(seconds=seconds))


class Container(object):
    def __init__(self, image, name,
                 volumes=None, environs=None,
                 label="", logger=None,
                 shared_memory="1gb",
                 time_out=-1,
                 log_container=None, 
                 workdir=None,
                 runscript=None):
        """
        Python wrapper to podman engine tools for managing containers.
        """

        self.image = image
        self.name = name
        self.label = label
        self.volumes = volumes or []
        self.environs = environs or []
        self.logger = logger
        self.status = None
        self.WORKDIR = workdir
        self.RUNSCRIPT = runscript
        self.shared_memory = shared_memory
        self.PID = os.getpid()
        self.uptime = "00:00:00"
        self.time_out = time_out
        self.cont_logger = utils.logger.StimelaLogger(
            log_container or stimela.LOG_FILE, jtype="podman")

    def add_volume(self, host, container, perm="rw", noverify=False):

        if os.path.exists(host) or noverify:
            if self.logger:
                self.logger.debug("Mounting volume [{0}] in container [{1}] at [{2}]".format(
                    host, self.name, container))
            host = os.path.abspath(host)
        else:
            raise IOError(
                "Directory {0} cannot be mounted on container: File doesn't exist".format(host))

        self.volumes.append(":".join([host, container, perm]))

    def add_environ(self, key, value):
        if self.logger:
            self.logger.debug("Adding environ varaible [{0}={1}] in container {2}".format(
                key, value, self.name))
        self.environs.append("=".join([key, value]))

    def create(self, *args):

        if self.volumes:
            volumes = " -v " + " -v ".join(self.volumes)
        else:
            volumes = ""
        if self.environs:
            environs = environs = " -e "+" -e ".join(self.environs)
        else:
            environs = ""

        # non-root podman users cannot allocate resources on all linux kernel setups. So minisise stress I don't
        # allow it
        args = list(args)
        for rsrc in ["--memory", "--cpus", "--user"]:
            for arg in args:
                if arg.startswith(rsrc):
                    args.remove(arg)

        self._print(
            "Instantiating container [{}]. The container ID is printed below.".format(self.name))
        utils.xrun("podman create", args + [volumes, environs, "--rm",
                                            "-w %s" % (self.WORKDIR) if self.WORKDIR else "",
                                            "--name", self.name, "--shm-size", self.shared_memory,
                                            self.image,
                                            self.RUNSCRIPT or ""], log=self.logger)

        self.status = "created"

    def info(self):

        output = subprocess.check_output(
            "podman inspect {}".format(self.name), shell=True).decode()
        output_file = StringIO(output[3:-3])
        jdict = json.load(output_file)
        output_file.close()

        return jdict

    def get_log(self):
        stdout = open(self.logfile, 'w')
        exit_status = subprocess.call("podman logs {0}".format(self.name),
                                      stdout=stdout, stderr=stdout, shell=True)
        if exit_status != 0:
            self.logger.warn(
                'Could not log container: {}. Something went wrong durring execution'.format(self.name))
            output = 'Task was not started.'
            stdout.write(output)
        else:
            output = stdout.read()

        stdout.close()
        return output

    def start(self):
        running = True
        tstart = time.time()
        self.status = "running"

        self.cont_logger.log_container(self.name)
        self.cont_logger.write()
        self._print("Starting container [{0:s}]. Timeout set to {1:d}. The container ID is printed below.".format(
            self.name, self.time_out))

        utils.xrun("podman", ["start", "-a", self.name],
                   timeout=self.time_out,
                   kill_callback=lambda: utils.xrun("podman", ["kill", self.name]),
                   logfile=self.logfile,
                   log=self.logger)

        uptime = seconds_hms(time.time() - tstart)
        self.uptime = uptime
        self._print(
            "Container [{0}] has executed successfully".format(self.name))

        self._print("Runtime was {0}.".format(uptime))

        self.status = "exited"

    def stop(self):
        dinfo = self.info()
        status = dinfo["State"]["Status"]
        killed = False
        if status in ["running", "paused"]:
            try:
                utils.xrun("podman stop", [self.name])
            except KeyboardInterrupt("Received terminate signal. Will stop and remove container first"):
                killed = True
        self.status = 'exited'

        self._print("Container {} has been stopped.".format(self.name))
        if killed:
            self.remove()
            raise KeyboardInterrupt

    def remove(self):
        dinfo = self.info()
        status = dinfo["State"]["Status"]
        killed = False
        if status == "exited":
            try:
                utils.xrun("podman rm", [self.name])
            except KeyboardInterrupt:
                killed = True
            if killed:
                raise KeyboardInterrupt

        else:
            raise DockerError(
                "Container [{}] has not been stopped, cannot remove".format(self.name))

        self.cont_logger.remove('containers', self.name)
        self.cont_logger.write()

    def _print(self, message):
        if self.logger:
            self.logger.info(message)
        else:
            print(message)



import os
import sys
import numpy
import Tigger
import tempfile
import pyfits
import shutil
import shlex
import subprocess
import yaml

from astLib.astWCS import WCS
from astropy.table import Table
from Tigger.Models import SkyModel, ModelClasses

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

with open(CONFIG, "r") as _std:
    cab = yaml.safe_load(_std)
junk = cab["junk"]

write_catalog = ['port2tigger', 'table']

write_opts = {}
img_opts = {}

args = []

for param in cab['parameters']:
    name = param['name']
    value = param['value']
    if value is None:
        continue
    elif value is False:
        continue
    if name == 'filename':  # positional argument
        args += ['{0}'.format(value)]
    else:
        if name != "port2tigger":
            args += ['{0}{1} {2}'.format(cab['prefix'], name, value)]
    if name in write_catalog:
        write_opts[name] = value
    else:
        img_opts[name] = value

_runc = " ".join([cab["binary"]] + args)

try:
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [OUTPUT, MSDIR]: # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)
                # Leave other types

port2tigger = write_opts.pop('port2tigger')
outfile = write_opts.pop('table')
image = img_opts.pop('filename')

if port2tigger:
    write_opts['format'] = 'fits'

if not port2tigger:
    sys.exit(0)


# convert to data file to Tigger LSM
# First make dummy tigger model
tfile = tempfile.NamedTemporaryFile(suffix='.txt')
tfile.flush()

prefix = os.path.splitext(outfile)[0]
tname_lsm = prefix + ".lsm.html"
with open(tfile.name, "w") as stdw:
    stdw.write("#format:name ra_d dec_d i emaj_s emin_s pa_d\n")

model = Tigger.load(tfile.name)
tfile.close()


def tigger_src(src, idx):

    name = "SRC%d" % idx
    flux = ModelClasses.Polarization(float(src["int_flux"]), 0, 0, 0,
                                     I_err=float(src["err_int_flux"]))
    ra, ra_err = map(numpy.deg2rad, (float(src["ra"]), float(src["err_ra"])))
    dec, dec_err = map(numpy.deg2rad, (float(src["dec"]),
                                       float(src["err_dec"])))
    pos = ModelClasses.Position(ra, dec, ra_err=ra_err, dec_err=dec_err)
    ex, ex_err = map(numpy.deg2rad, (float(src["a"]), float(src["err_a"])))
    ey, ey_err = map(numpy.deg2rad, (float(src["b"]), float(src["err_b"])))
    pa, pa_err = map(numpy.deg2rad, (float(src["pa"]), float(src["err_pa"])))

    if ex and ey:
        shape = ModelClasses.Gaussian(
            ex, ey, pa, ex_err=ex_err, ey_err=ey_err, pa_err=pa_err)
    else:
        shape = None
    source = SkyModel.Source(name, pos, flux, shape=shape)
    # Adding source peak flux (error) as extra flux attributes for sources,
    # and to avoid null values for point sources I_peak = src["Total_flux"]
    if shape:
        source.setAttribute("I_peak", float(src["peak_flux"]))
        source.setAttribute("I_peak_err", float(src["err_peak_flux"]))
    else:
        source.setAttribute("I_peak", float(src["int_flux"]))
        source.setAttribute("I_peak_err", float(src["err_int_flux"]))

    return source


data = Table.read('{0}_comp.{1}'.format(outfile.split('.')[0], outfile.split('.')[-1]),
                  format=outfile.split('.')[-1])

for i, src in enumerate(data):

    model.sources.append(tigger_src(src, i))

wcs = WCS(image)
centre = wcs.getCentreWCSCoords()
model.ra0, model.dec0 = map(numpy.deg2rad, centre)

model.save(tname_lsm)
# Rename using CORPAT
utils.xrun("tigger-convert", [tname_lsm, "--rename -f"])



# -*- coding: future_fstrings -*-
class StimelaCabParameterError(Exception):
    pass

class StimelaRecipeExecutionError(Exception):
    pass

class PipelineException(Exception):
    """ 
    Encapsulates information about state of pipeline when an
    exception occurs
    """

    def __init__(self, exception, completed, failed, remaining):
        message = ("Job '%s' failed: %s" % (failed.label, str(exception)))

        super(PipelineException, self).__init__(message)

        self._completed = completed
        self._failed = failed
        self._remaining = remaining

    @property
    def completed(self):
        return self._completed

    @property
    def failed(self):
        return self._failed

    @property
    def remaining(self):
        return self._remaining




import os

ekhaya = os.path.dirname(__file__)

# Path to base images
BASE_PATH = "{:s}/base".format(ekhaya)

# Path to executor images
CAB_PATH = "{:s}/cab".format(ekhaya)

# Path to config templates
CONFIG_TEMPLATES = "{:s}/configs".format(ekhaya)



import os
import sys
import shlex
import shutil
import subprocess
import yaml 
import glob

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

with open(CONFIG, "r") as _std:
    cab = yaml.safe_load(_std)

junk = cab["junk"]

args = []
for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if name == 'ms':
        mslist = ' '.join(value)
        continue
    if value in [None, False]:
        continue
    if value is True:
        value = ""
    args += ['{0}{1} {2}'.format(cab['prefix'], name, value)]

_runc = " ".join(["mask_ms.py"] + args + [mslist])
try:
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [OUTPUT, MSDIR]: # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)



import os
import sys
import glob
import subprocess
import yaml
import shutil
import shlex

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

with open(CONFIG, "r") as _std:
    cab = yaml.safe_load(_std)

junk = cab["junk"]

params = {}
for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if value in [False, None]:
        continue

    if value is True:
        value = ""

    # Positional arguments
    if name == 'input-skymodel':
        inlsm = value
        continue
    elif name == 'tag':
        tag = value
        continue

    params[name] = value

# TODO: Need fix tigger-tag, these kludges are annoying
if params.pop('transfer-tags', False) in [True, ""]:
    if params.get('tolerance', None) is None:
        raise RuntimeError(
            "Parameter 'tolerance' is required when 'transfer-tags' is enables")
    args = [
        '{0}transfer-tags {1}:{2}'.format(cab['prefix'], inlsm, params.pop('tolerance'))]
    inlsm = params.get('output')
else:
    args = []

args += ['{0}{1} {2}'.format(cab['prefix'], name, value)
         for name, value in params.iteritems()]

_runc = " ".join([cab['binary']] args+[inlsm, tag])

try:
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [OUTPUT, MSDIR]: # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)



import os
import sys
import shlex
import shutil
import yaml
import glob
import subprocess

OUTPUT = os.environ["OUTPUT"]
CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]

with open(CONFIG, "r") as _std:
    cab = yaml.safe_load(_std)
junk = cab["junk"]

params = cab["parameters"]

args = []
for param in params:
    if param['value'] in [False, None]:
        continue
    elif param['value'] is True:
        arg = "{0}{1}".format(cab["prefix"], param["name"])
    else:
        arg = "{0}{1} {2}".format(cab["prefix"], param["name"], param["value"])

    args.append(arg)

_runc = " ".join([cab["binary"]] + args)

try:
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [OUTPUT, MSDIR]: # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)




import os
import sys
import json
import yaml
import time
import tempfile
import inspect
import warnings
import re
import math
import codecs

class StimelaCabRuntimeError(RuntimeError):
    pass

class StimelaProcessRuntimeError(RuntimeError):
    pass



CPUS = 1

from .xrun_poll import xrun

def assign(key, value):
    frame = inspect.currentframe().f_back
    frame.f_globals[key] = value


def readJson(conf):
    with open(conf, "r") as _std:
        jdict = yaml.safe_load(_std)
        return jdict


def writeJson(config, dictionary):
    with codecs.open(config, 'w', 'utf8') as std:
        std.write(json.dumps(dictionary, ensure_ascii=False))


def get_Dockerfile_base_image(image):

    if os.path.isfile(image):
        dockerfile = image
    else:
        dockerfile = "{:s}/Dockerfile".format(image)

    with open(dockerfile, "r") as std:
        _from = ""
        for line in std.readlines():
            if line.startswith("FROM"):
                _from = line

    return _from


def change_Dockerfile_base_image(path, _from, label, destdir="."):
    if os.path.isfile(path):
        dockerfile = path
        dirname = os.path.dirname(path)
    else:
        dockerfile = "{:s}/Dockerfile".format(path)
        dirname = path

    with open(dockerfile, "r") as std:
        lines = std.readlines()
        for line in lines:
            if line.startswith("FROM"):
                lines.remove(line)

    temp_dir = tempfile.mkdtemp(
        prefix="tmp-stimela-{:s}-".format(label), dir=destdir)
    xrun(
        "cp", ["-r", "{:s}/Dockerfile {:s}/src".format(dirname, dirname), temp_dir])

    dockerfile = "{:s}/Dockerfile".format(temp_dir)

    with open(dockerfile, "w") as std:
        std.write("{:s}\n".format(_from))

        for line in lines:
            std.write(line)

    return temp_dir, dockerfile


def get_base_images(logfile, index=1):

    with opEn(logfile, "r") as std:
        string = std.read()

    separator = "[================================DONE==========================]"

    log = string.split(separator)[index-1]

    images = []

    for line in log.split("\n"):
        if line.find("<=BASE_IMAGE=>") > 0:
            tmp = line.split("<=BASE_IMAGE=>")[-1]
            image, base = tmp.split("=")
            images.append((image.strip(), base))

    return images


def icasa(taskname, mult=None, clearstart=False, loadthese=[], **kw0):
    """ 
      runs a CASA task given a list of options.
      A given task can be run multiple times with a different options, 
      in this case the options must be parsed as a list/tuple of dictionaries via mult, e.g 
      icasa('exportfits',mult=[{'imagename':'img1.image','fitsimage':'image1.fits},{'imagename':'img2.image','fitsimage':'image2.fits}]). 
      Options you want be common between the multiple commands should be specified as key word args.
    """

    # create temp directory from which to run casapy
    td = tempfile.mkdtemp(dir='.')
    # we want get back to the working directory once casapy is launched
    cdir = os.path.realpath('.')

    # load modules in loadthese
    _load = ""
    if "os" not in loadthese or "import os" not in loadthese:
        loadthese.append("os")

    if loadthese:
        exclude = filter(lambda line: line.startswith("import")
                         or line.startswith("from"), loadthese)
        for line in loadthese:
            if line not in exclude:
                line = "import %s" % line
            _load += "%s\n" % line

    if mult:
        if isinstance(mult, (tuple, list)):
            for opts in mult:
                opts.update(kw0)
        else:
            mult.upadte(kw0)
            mult = [mult]
    else:
        mult = [kw0]

    run_cmd = """ """
    for kw in mult:
        task_cmds = []
        for key, val in kw.items():
            if isinstance(val, (str, unicode)):
                val = '"%s"' % val
            task_cmds .append('%s=%s' % (key, val))

        task_cmds = ", ".join(task_cmds)
        run_cmd += """ 
%s
os.chdir('%s')
%s
%s(%s)
""" % (_load, cdir, "clearstart()" if clearstart else "", taskname, task_cmds)

    tf = tempfile.NamedTemporaryFile(suffix='.py')
    tf.write(run_cmd)
    tf.flush()
    t0 = time.time()
    # all logging information will be in the pyxis log files
    print("Running {}".format(run_cmd))
    xrun("cd", [td, "&& casa --nologger --log2term --nologfile -c", tf.name])

    # log taskname.last
    task_last = '%s.last' % taskname
    if os.path.exists(task_last):
        with opEn(task_last, 'r') as last:
            print('%s.last is: \n %s' % (taskname, last.read()))

    # remove temp directory. This also gets rid of the casa log files; so long suckers!
    xrun("rm", ["-fr ", td, task_last])
    tf.close()


def stack_fits(fitslist, outname, axis=0, ctype=None, keep_old=False, fits=False):
    """ Stack a list of fits files along a given axiis.

       fitslist: list of fits file to combine
       outname: output file name
       axis: axis along which to combine the files
       fits: If True will axis FITS ordering axes
       ctype: Axis label in the fits header (if given, axis will be ignored)
       keep_old: Keep component files after combining?
    """
    import numpy
    try:
        import pyfits
    except ImportError:
        warnings.warn(
            "Could not find pyfits on this system. FITS files will not be stacked")
        sys.exit(0)

    hdu = pyfits.open(fitslist[0])[0]
    hdr = hdu.header
    naxis = hdr['NAXIS']

    # find axis via CTYPE key
    if ctype is not None:
        for i in range(1, naxis+1):
            if hdr['CTYPE%d' % i].lower().startswith(ctype.lower()):
                axis = naxis - i  # fits to numpy convention
    elif fits:
        axis = naxis - axis

    fits_ind = abs(axis-naxis)
    crval = hdr['CRVAL%d' % fits_ind]

    imslice = [slice(None)]*naxis
    _sorted = sorted([pyfits.open(fits) for fits in fitslist],
                     key=lambda a: a[0].header['CRVAL%d' % (naxis-axis)])

    # define structure of new FITS file
    nn = [hd[0].header['NAXIS%d' % (naxis-axis)] for hd in _sorted]
    shape = list(hdu.data.shape)
    shape[axis] = sum(nn)
    data = numpy.zeros(shape, dtype=float)

    for i, hdu0 in enumerate(_sorted):
        h = hdu0[0].header
        d = hdu0[0].data
        imslice[axis] = range(sum(nn[:i]), sum(nn[:i+1]))
        data[imslice] = d
        if crval > h['CRVAL%d' % fits_ind]:
            crval = h['CRVAL%d' % fits_ind]

    # update header
    hdr['CRVAL%d' % fits_ind] = crval
    hdr['CRPIX%d' % fits_ind] = 1

    pyfits.writeto(outname, data, hdr, clobber=True)
    print("Successfully stacked images. Output image is %s" % outname)

    # remove old files
    if not keep_old:
        for fits in fitslist:
            os.system('rm -f %s' % fits)


def substitute_globals(string, globs=None):
    sub = set(re.findall('\{(.*?)\}', string))
    globs = globs or inspect.currentframe().f_back.f_globals
    if sub:
        for item in map(str, sub):
            string = string.replace("${%s}" % item, globs[item])
        return string
    else:
        return False


def get_imslice(ndim):
    imslice = []
    for i in xrange(ndim):
        if i < ndim-2:
            imslice.append(0)
        else:
            imslice.append(slice(None))

    return imslice


def addcol(msname, colname=None, shape=None,
           data_desc_type='array', valuetype=None, init_with=0, **kw):
    """ add column to MS 
        msanme : MS to add colmn to
        colname : column name
        shape : shape
        valuetype : data type 
        data_desc_type : 'scalar' for scalar elements and array for 'array' elements
        init_with : value to initialise the column with 
    """
    import numpy
    import pyrap.tables
    tab = pyrap.tables.table(msname, readonly=False)

    try:
        tab.getcol(colname)
        print('Column already exists')

    except RuntimeError:
        print('Attempting to add %s column to %s' % (colname, msname))

        from pyrap.tables import maketabdesc
        valuetype = valuetype or 'complex'

        if shape is None:
            dshape = list(tab.getcol('DATA').shape)
            shape = dshape[1:]

        if data_desc_type == 'array':
            from pyrap.tables import makearrcoldesc
            # God forbid this (or the TIME) column doesn't exist
            coldmi = tab.getdminfo('DATA')
            coldmi['NAME'] = colname.lower()
            tab.addcols(maketabdesc(makearrcoldesc(
                colname, init_with, shape=shape, valuetype=valuetype)), coldmi)

        elif data_desc_type == 'scalar':
            from pyrap.tables import makescacoldesc
            coldmi = tab.getdminfo('TIME')
            coldmi['NAME'] = colname.lower()
            tab.addcols(maketabdesc(makescacoldesc(
                colname, init_with, valuetype=valuetype)), coldmi)

        print('Column added successfuly.')

        if init_with:
            nrows = dshape[0]

            rowchunk = nrows//10 if nrows > 1000 else nrows
            for row0 in range(0, nrows, rowchunk):
                nr = min(rowchunk, nrows-row0)
                dshape[0] = nr
                tab.putcol(colname, numpy.ones(
                    dshape, dtype=valuetype)*init_with, row0, nr)

    tab.close()


def sumcols(msname, col1=None, col2=None, outcol=None, cols=None, suntract=False):
    """ add col1 to col2, or sum columns in 'cols' list.
        If subtract, subtract col2 from col1
    """
    from pyrap.tables import table

    tab = table(msname, readonly=False)
    if cols:
        data = 0
        for col in cols:
            data += tab.getcol(col)
    else:
        if subtract:
            data = tab.getcol(col1) - tab.getcol(col2)
        else:
            data = tab.getcol(col1) + tab.getcol(col2)

    rowchunk = nrows//10 if nrows > 1000 else nrows
    for row0 in range(0, nrows, rowchunk):
        nr = min(rowchunk, nrows-row0)
        tab.putcol(outcol, data[row0:row0+nr], row0, nr)

    tab.close()


def copycol(msname, fromcol, tocol):
    from pyrap.tables import table

    tab = table(msname, readonly=False)
    data = tab.getcol(fromcol)
    if tocol not in tab.colnames():
        addcol(msname, tocol)

    nrows = tab.nrows()
    rowchunk = nrows//10 if nrows > 5000 else nrows
    for row0 in range(0, nrows, rowchunk):
        nr = min(rowchunk, nrows-row0)
        tab.putcol(tocol, data[row0:row0+nr], row0, nr)

    tab.close()


def cab_dict_update(dictionary, key=None, value=None, options=None):
    if options is None:
        options = {key: value}
    for key, value in options.items():
        dictionary[key] = dictionary.pop(key, None) or value
    return dictionary


def compute_vis_noise(msname, sefd, spw_id=0):
    """Computes nominal per-visibility noise"""
    from pyrap.tables import table

    tab = table(msname)
    spwtab = table(msname + "/SPECTRAL_WINDOW")

    freq0 = spwtab.getcol("CHAN_FREQ")[spw_id, 0]
    wavelength = 300e+6/freq0
    bw = spwtab.getcol("CHAN_WIDTH")[spw_id, 0]
    dt = tab.getcol("EXPOSURE", 0, 1)[0]
    dtf = (tab.getcol("TIME", tab.nrows()-1, 1)-tab.getcol("TIME", 0, 1))[0]

    # close tables properly, else the calls below will hang waiting for a lock...
    tab.close()
    spwtab.close()

    print(">>> %s freq %.2f MHz (lambda=%.2fm), bandwidth %.2g kHz, %.2fs integrations, %.2fh synthesis" % (
        msname, freq0*1e-6, wavelength, bw*1e-3, dt, dtf/3600))
    noise = sefd/math.sqrt(abs(2*bw*dt))
    print(">>> SEFD of %.2f Jy gives per-visibility noise of %.2f mJy" %
          (sefd, noise*1000))

    return noise



import sys, os, codecs, time, hashlib, subprocess

from threading import Event, Thread

from . import StimelaCabRuntimeError

DEBUG = False
INTERRUPT_TIME = 2.0  # seconds -- do not want to constantly interrupt the child process
LIVELOG_TIME = 0.1



def xrun(command, options, log=None, _log_container_as_started=False, logfile=None, timeout=-1, kill_callback=None):
    """
        Run something on command line.

        Example: _run("ls", ["-lrt", "../"])
    """
    if "LOGFILE" in os.environ and logfile is None:
        logfile = os.environ["LOGFILE"] # superceed if not set

    # skip lines from previous log files
    if logfile is not None and os.path.exists(logfile):
        with codecs.open(logfile, "r", encoding="UTF-8",
                         errors="ignore", buffering=0) as foutlog:
            lines = foutlog.readlines()
            prior_log_bytes_read = foutlog.tell()
    else: # not existant, create
        prior_log_bytes_read = 0
        if logfile is not None and not os.path.exists(logfile):
            with codecs.open(logfile, "w+", encoding="UTF-8",
                             errors="ignore", buffering=0) as foutlog:
                pass

    cmd = " ".join([command] + list(map(str, options)))

    def _remove_ctrls(msg):
        import re
        ansi_escape = re.compile(r'\x1B[@-_][0-?]*[ -/]*[@-~]')
        return ansi_escape.sub('', msg)

    def _print_info(msg, loglevel="INFO"):
        if msg is None:
            return
        msg = _remove_ctrls(msg)
        if msg.strip() == "": return
        if log:
            try:
                getattr(log, loglevel.lower())(msg.rstrip('\n'))
            except UnicodeError:
                log.warn("Log contains unicode and will not be printed")
        else:
            try:
                print(msg),
            except UnicodeError:
                print("Log contains unicode and will not be printed")

    def _print_warn(msg):
        if msg is None:
            return
        msg = _remove_ctrls(msg)
        if msg.strip() == "": return
        if log:
            try:
                log.info(msg.rstrip('\n'))
            except UnicodeError:
                log.warn("Log contains unicode and will not be printed")
        else:
            try:
                print(msg),
            except UnicodeError:
                print("Log contains unicode and will not be printed")


    _print_info(u"Running: {0:s}".format(cmd), loglevel="INFO")

    sys.stdout.flush()
    starttime = time.time()
    process = p = None
    stop_log_printer = Event()

    try:
        foutname = os.path.join("/tmp", "stimela_output_{0:s}_{1:f}".format(
            hashlib.md5(cmd.encode('utf-8')).hexdigest(), starttime))


        p = process = subprocess.Popen(cmd,
                                       shell=True)
        kill_callback = kill_callback or p.kill

        def clock_killer(p):
            while process.poll() is None and (timeout >= 0):
                currenttime = time.time()
                if (currenttime - starttime < timeout):
                    DEBUG and _print_warn(u"Clock Reaper: has been running for {0:f}, must finish in {1:f}".format(
                        currenttime - starttime, timeout))
                else:
                    _print_warn(
                        u"Clock Reaper: Timeout reached for '{0:s}'... sending the KILL signal".format(cmd))
                    kill_callback()
                time.sleep(INTERRUPT_TIME)

        def log_reader(logfile, stop_event):
            bytes_read = prior_log_bytes_read # skip any previous runs' output
            while not stop_event.isSet():
                if logfile is not None and os.path.exists(logfile):
                    with codecs.open(logfile, "r", encoding="UTF-8",
                                     errors="ignore", buffering=0) as foutlog:
                        foutlog.seek(bytes_read, 0)
                        lines = foutlog.readlines()
                        bytes_read = foutlog.tell()
                        for line in lines:
                            line and _print_info(line)
                time.sleep(LIVELOG_TIME) # wait for the log to go to disk

        Thread(target=clock_killer, args=tuple([p])).start()
        if log is not None:
            # crucial - child process should not write to stdout unless it is
            # the container process itself
            Thread(target=log_reader, args=tuple([logfile, stop_log_printer])).start()


        while (process.poll() is None):
            currenttime = time.time()
            DEBUG and _print_info(
                u"God mode on: has been running for {0:f}".format(currenttime - starttime))
            # this is probably not ideal as it interrupts the process every few seconds,
            time.sleep(INTERRUPT_TIME)
            # check whether there is an alternative with a callback

        assert hasattr(
            process, "returncode"), "No returncode after termination!"
    finally:
        stop_log_printer.set()
        if (process is not None) and process.returncode:
            raise StimelaCabRuntimeError(
                '%s: returns errr code %d' % (command, process.returncode))




import select, traceback, subprocess, errno, re, time, logging, os, sys

DEBUG = 0
from . import StimelaCabRuntimeError, StimelaProcessRuntimeError

log = None

def get_stimela_logger():
    """Returns Stimela's logger, or None if no Stimela installed"""
    try:
        import stimela
        return stimela.logger()
    except ImportError:
        return None

def global_logger():
    """Returns Stimela logger if running in stimela, else inits a global logger"""
    global log
    if log is None:
        log = get_stimela_logger()
        if log is None:
            # no stimela => running payload inside a cab -- just use the global logger and make it echo everything to the console
            logging.basicConfig(format="%(message)s", level=logging.INFO, stream=sys.stdout)
            log = logging.getLogger()
    return log

class Poller(object):
    """Poller class. Poor man's select.poll(). Damn you OS/X and your select.poll will-you-won'y-you bollocks"""
    def __init__ (self, log):
        self.fdlabels = {}
        self.log = log

    def register_file(self, fobj, label):
        self.fdlabels[fobj.fileno()] = label, fobj

    def register_process(self, po, label_stdout='stdout', label_stderr='stderr'):
        self.fdlabels[po.stdout.fileno()] = label_stdout, po.stdout
        self.fdlabels[po.stderr.fileno()] = label_stderr, po.stderr

    def poll(self, timeout=5, verbose=False):
        while True:
            try:
                to_read, _, _ = select.select(self.fdlabels.keys(), [], [], timeout)
                self.log.debug("poll(): ready to read: {}".format(to_read))
                # return on success or timeout
                return [self.fdlabels[fd] for fd in to_read]
            except (select.error, IOError) as ioerr:
                if verbose:
                    self.log.debug("poll() exception: {}".format(traceback.format_exc()))
                if hasattr(ioerr, 'args'):
                    err = ioerr.args[0]  # py2
                else:
                    err = ioerr.errno    # py3
                # catch interrupted system call -- return if we have a timeout, else
                # loop again
                if err == errno.EINTR:
                    if timeout is not None:
                        if verbose:
                            self.log.debug("poll(): returning")
                        return []
                    if verbose:
                        self.log.debug("poll(): retrying")
                else:
                    raise ioerr

    def unregister_file(self, fobj):
        if fobj.fileno() in self.fdlabels:
            del self.fdlabels[fobj.fileno()]

    def __contains__(self, fobj):
        return fobj.fileno() in self.fdlabels


def _remove_ctrls(msg):
    ansi_escape = re.compile(r'\x1B[@-_][0-?]*[ -/]*[@-~]')
    return ansi_escape.sub('', msg)


def xrun_nolog(command, name=None):
    log = global_logger()
    name = name or command.split(" ", 1)[0]
    try:
        log.info("# running {}".format(command))
        status = subprocess.call(command, shell=True)

    except KeyboardInterrupt:
        log.error("# {} interrupted by Ctrl+C".format(name))
        raise

    except Exception as exc:
        for line in traceback.format_exc():
            log.error("# {}".format(line.strip()))
        log.error("# {} raised exception: {}".format(name, str(exc)))
        raise

    if status:
        raise StimelaProcessRuntimeError("{} returns error code {}".format(name, status))

    return 0

def xrun(command, options, log=None, logfile=None, timeout=-1, kill_callback=None):
    command_name = command

    # this part could be inside the container
    command = " ".join([command] + list(map(str, options)))

    log = log or get_stimela_logger()

    if log is None:
        return xrun_nolog(command, name=command_name)

    # this part is never inside the container
    import stimela

    log = log or stimela.logger()

    log.info("running " + command, extra=dict(stimela_subprocess_output=(command_name, "start")))

    start_time = time.time()

    proc = subprocess.Popen([command], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                            bufsize=1, universal_newlines=True, shell=True)

    poller = Poller(log=log)
    poller.register_process(proc)

    proc_running = True

    try:
        while proc_running and poller.fdlabels:
            fdlist = poller.poll(verbose=DEBUG>0)
            for fname, fobj in fdlist:
                try:
                    line = fobj.readline()
                except EOFError:
                    line = b''
                empty_line = not line
                line = (line.decode('utf-8') if type(line) is bytes else line).rstrip()
                # break out if process closes
                if empty_line:
                    poller.unregister_file(fobj)
                    if proc.stdout not in poller and proc.stderr not in poller:
                        log.debug("The {} process has exited".format(command))
                        proc_running = None
                        break
                    continue
                # dispatch output to log
                line = _remove_ctrls(line)
                # the extra attributes are filtered by e.g. the CARACal logger
                if fobj is proc.stderr:
                    log.warning(line, extra=dict(stimela_subprocess_output=(command_name, "stderr")))
                else:
                    log.info(line, extra=dict(stimela_subprocess_output=(command_name, "stdout")))
            if timeout > 0 and time.time() > start_time + timeout:
                log.error("timeout, killing {} process".format(command))
                kill_callback() if callable(kill_callback) else proc.kill()
                proc_running = False

        proc.wait()
        status = proc.returncode

    except SystemExit as exc:
        log.error("{} has exited with code {}".format(command, exc.code))
        proc.wait()
        status = exc.code
        raise StimelaCabRuntimeError('{}: SystemExit with code {}'.format(command_name, status))

    except KeyboardInterrupt:
        if callable(kill_callback):
            log.warning("Ctrl+C caught: shutting down {} process, please give it a few moments".format(command_name))
            kill_callback() 
            log.info("the {} process was shut down successfully".format(command_name),
                     extra=dict(stimela_subprocess_output=(command_name, "status")))
        else:
            log.warning("Ctrl+C caught, killing {} process".format(command_name))
            proc.kill()
        proc.wait()
        raise StimelaCabRuntimeError('{} interrupted with Ctrl+C'.format(command_name))

    except Exception as exc:
        traceback.print_exc()
        log.error("Exception caught: {}".format(str(exc)))
        proc.wait()
        raise StimelaCabRuntimeError("{} throws exception '{}'".format(command_name, str(exc)))

    if status:
        raise StimelaCabRuntimeError("{} returns error code {}".format(command_name, status))
    
    return status
    



import os
import sys
import subprocess
import shlex
import shutil
import glob
import yaml

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
MSDIR = os.environ["MSDIR"]
OUTPUT = os.environ["OUTPUT"]

with open(CONFIG, "r") as _std:
    cab = yaml.safe_load(_std)
junk = cab["junk"]

args = {}
for param in cab['parameters']:
    name = param['name']
    value = param['value']

    if value is None:
        continue
    args[name] = value

if os.path.exists(OUTPUT+'/mask_mosaic') == False:
    os.mkdir(OUTPUT+'/mask_mosaic')

outdir = OUTPUT+'/mask_mosaic'

try:
    make_table = " ".join(['mImgtbl', args['input_dir'], outdir+'/mosaic_table.tbl'])
    subprocess.check_call(shlex.split(make_table))

    make_header = " ".join(['mMakeHdr', outdir +
               '/mosaic_table.tbl', outdir+'/mosaic_header.hdr'])
    subprocess.check_call(shlex.split(make_header))

    project_mosaic = " ".join(['mProjExec', '-p', args['input_dir'], outdir +
                  '/mosaic_table.tbl', outdir+'/mosaic_header.hdr', outdir, outdir+'/stats.tbl'])
    subprocess.check_call(shlex.split(project_mosaic))

    make_mosaic_table = ['mImgtbl', outdir, outdir+'/mosaic_table2.tbl']
    subprocess.check_call(shlex.split(make_mosaic_table))

    _runc = " ".join(['mAdd', '-p', args['input_dir'], outdir +
               '/mosaic_table2.tbl', outdir+'/mosaic_header.hdr', OUTPUT+'/mosaic.fits'])
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [OUTPUT, MSDIR]: # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)
                # Leave other types
